/*
 * SourceViewer.java
 *
 * Created on July 9, 2001, 7:39 AM
 *
 * See LICENSE file for license conditions.
 */

package residue;
import javax.swing.*;
import java.util.*;
import java.io.*;
import java.awt.Color;
import java.awt.Component;
import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.Rectangle;
import java.awt.Toolkit;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseListener;
import java.awt.event.MouseEvent;
import java.awt.image.MemoryImageSource;
import java.awt.Image;

import residue.tables.*;

/**
 * This class is the simple viewer that highlights lines
 * to indicate their execution status in the hit table.
 * 
 * Its implementation is kind of a hack.  It is a JList
 * with a new ListCellRenderer, rather than a complex
 * design with a JEditorPane.
 * 
 * @author  chowells
 */
public class SourceViewer extends javax.swing.JFrame
{
	private boolean quit;
	private boolean stored;
	private String fname;
	
	private Rectangle correctBounds;
	
	private String packagesource;
	
	private Table lt;
	
	// line (Integer) -> Set of LineTableEntry map
	private Map entries;
	
	// Set of lines (Integer) that were hit
	private Set hits;
	
	// Set of lines (Integer) that were missed
	private Set misses;
	
    // JList containing line contents
    private JList lineData;
    
    // JList containing line numbers
    private JList lineNumbers;
    
    // ListCellRenderer for standard view
    private ListCellRenderer highlightingRenderer = new HighlightingListCellRenderer();
    
    // ListCellRenderer for summary view
    private ListCellRenderer summaryRenderer = new SummaryListCellRenderer();
    
	// property names
	private static final String EXECUTED_COLOR_PROPERTY = "sourceviewer.executedColor";
	private static final String NOTEXECUTED_COLOR_PROPERTY = "sourceviewer.notExColor";
	private static final String MIXED_COLOR_PROPERTY = "sourceviewer.mixedColor";
	private static final String TOP_BOUND_PROPERTY = "sourceviewer.top";
	private static final String LEFT_BOUND_PROPERTY = "sourceviewer.left";
	private static final String WIDTH_PROPERTY = "sourceviewer.width";
	private static final String HEIGHT_PROPERTY = "sourceviewer.height";
	
	private static Color exec;
	private static Color not;
	private static Color mix;
	
	private static JFileChooser fc = new JFileChooser();
	
	static
	{
		fc.setCurrentDirectory(new File(System.getProperty("user.dir")));
		
		String propname = System.getProperty("user.home") + File.separator + ".gretelrc";
		
		Properties p = new Properties();
		try
		{
			InputStream is = new FileInputStream(propname);
			p.load(is);
			
			is.close();
		}
		catch (IOException e) {}
		
		try { exec = Color.decode(p.getProperty(EXECUTED_COLOR_PROPERTY)); }
		catch (Exception e) { exec = new Color(0x66ff66); }
			
		try { not = Color.decode(p.getProperty(NOTEXECUTED_COLOR_PROPERTY)); }
		catch (Exception e) { not = new Color(0xff0000); }
		
		try { mix = Color.decode(p.getProperty(MIXED_COLOR_PROPERTY)); }
		catch (Exception e) { mix = new Color(0xffcc33); }
	} // static
	
	
	public static void main(String [] args) throws IOException
	{
		ListHits.setConfigFile(args[0]);
		SourceViewer s = new SourceViewer(args[1], true);
		s.show();
		s.fixBounds();
	} // main
	
	/**
	 * Creates new SourceViewer for the .java file specified in
	 * the first argument.  If quitOnExit is true, the JVM will
	 * be shut down (System.exit(0) will be called) when the
	 * form is closed.  Otherwise, the frame will just call
	 * dispose() on itself.
	 */
	public SourceViewer(String filename, boolean quitOnExit) throws IOException
	{
		quit = quitOnExit;
		packagesource = filename;
		resolvefname(filename);
		
		initComponents();
		fill();
		
		setTitle(fname);
		pack();
		loadLocation();
	} // SourceViewer
	
	
	/**
	 * sets the bounds of this window according to the contents of
	 * the properties file.
	 */
	private void loadLocation()
	{
		Dimension d = Toolkit.getDefaultToolkit().getScreenSize();
		
		int top = (d.height - getHeight()) / 2;
		int left = (d.width - getWidth()) / 2;
		int w = getWidth();
		int h = getHeight();
		
		String propname = System.getProperty("user.home") + File.separator + ".gretelrc";
		
		Properties p = new Properties();
		try
		{
			InputStream is = new FileInputStream(propname);
			p.load(is);
			
			is.close();
		}
		catch (IOException e) {}
		
		try { top = Integer.decode(p.getProperty(TOP_BOUND_PROPERTY)).intValue(); }
		catch (Exception e) {}
		if (top > d.height) top = d.height - 5;
		
		try { left = Integer.decode(p.getProperty(LEFT_BOUND_PROPERTY)).intValue(); }
		catch (Exception e) {}
		if (left > d.width) top = d.width - 5;
		
		try { w = Integer.decode(p.getProperty(WIDTH_PROPERTY)).intValue(); }
		catch (Exception e) {}
		
		try { h = Integer.decode(p.getProperty(HEIGHT_PROPERTY)).intValue(); }
		catch (Exception e) {}
		
		setBounds(left, top, w, h);
		correctBounds = new Rectangle(left, top, w, h);
	} // loadLocation
	
	/**
	 * Fills in the entries Map.
	 */
	private void fillTables()
	{
		entries = new HashMap();
		hits = new HashSet();
		misses = new HashSet();
		
		try
		{
			lt = ListHits.getTable();
		}
		catch (IOException e)
		{
			return;
		}
		
		Iterator lines = lt.iterator();
		while (lines.hasNext())
		{
			TableEntry te = (TableEntry)lines.next();
			
			// make sure that only lines related to this source file
			// get included
			if (!te.getSourceName().equals(packagesource)) continue;
			
			boolean executed = te.wasExecuted();
			
			Iterator it = te.getLineSet().iterator();
			while (it.hasNext())
			{
				Object key = it.next();
				if (!entries.containsKey(key))
					entries.put(key, new HashSet());
				
				Set s = (Set)entries.get(key);
				s.add(te);
				
				if (executed) hits.add(key);
				else misses.add(key);
				
			} // while
		} // while
		
	} // fillEntries
	
	/**
	 * This method attempts to convert a filename consisting of
	 * a package and source name into an absolute filename.
	 *
	 * First, it checks all of the roots list.  If it can't
	 * find the file using any of them, it asks the user where
	 * it is.
	 */
	private void resolvefname(String filename) throws IOException
	{
		try
		{
			lt = ListHits.getTable();
		}
		catch (IOException e)
		{
			JOptionPane.showMessageDialog(this, "Unable to locate the line table.",
				"Error", JOptionPane.ERROR_MESSAGE);
			throw e;
		}
		
		Set s = lt.getSourceRoots();
		
		Iterator it = s.iterator();
		while (it.hasNext())
		{
			String root = (String)it.next();
			
			File rootdir = new File(root);
			
			File maybe = new File(rootdir.getAbsolutePath() + File.separator + filename);
			
			if (maybe.exists())
			{
				fname = maybe.getAbsolutePath();
				return;
			}
		} // while
		
		fc.setDialogTitle("Please locate the source for " + filename);
		
		int result = fc.showOpenDialog(this);
		if (result != fc.APPROVE_OPTION)
			throw new IOException("File " + filename + " not found.");
		
		fname = fc.getSelectedFile().getAbsolutePath();

		try
		{
			String newroot = fname.substring(0, fname.lastIndexOf(packagesource));
			lt.addSourceRoot(newroot, residue.runtime.Monitor.getLineTableName());
		}
		catch (IndexOutOfBoundsException e)
		{
			JOptionPane.showMessageDialog(this,
				"Source file has the wrong name.\nSee help for more information",
				"Error", JOptionPane.ERROR_MESSAGE);
		} // catch
	} // resolvefname
	
	/**
	 * Actually reads in the file and the necessary tables to 
	 * fill in the source code, and highlight it appropriately.
	 */
	private void fill() throws IOException
	{
		Vector v = new Vector();
		Vector lines = new Vector();
		
		try
		{
			ListHits.update();
		} catch (IOException e) {}
		
		BufferedReader r = new BufferedReader(new FileReader(fname));
		
		fillTables();
		
		int i = 0;
		final boolean linezeroshown;
		
		// set up line 0 in the SourceViewer if needed
		if (hits.contains(new Integer(0)) || misses.contains(new Integer(0)))
		{
			HighlightedLine hl = new HighlightedLine();
			hl.text = "// This is line 0, inserted because some code claims line 0 as its source.";
			
			if (hits.contains(new Integer(0))) hl.attribs |= 1;
			if (misses.contains(new Integer(0))) hl.attribs |= 2;
			
			v.add(hl);
			
			HighlightedLine l = new HighlightedLine();
			l.text = "0";
			
			if (hits.contains(new Integer(0))) l.attribs |= 1;
			if (misses.contains(new Integer(0))) l.attribs |= 2;
			
			lines.add(l);
			linezeroshown = true;
		}
		else
		{
			linezeroshown = false;
		}
		
		// loop through source, setting up HighlightedLines
		while (true)
		{
			String s = r.readLine();
			if (s == null) break;
			i++;
			
			StringBuffer temp = new StringBuffer();
			for (int j = 0; j < s.length(); j++)
			{
				if (s.charAt(j) == '\t') temp.append("        ");
				else temp.append(s.charAt(j));
			}
			temp.append(" ");
			s = temp.toString();
			
			HighlightedLine hl = new HighlightedLine();
			hl.text = s;

			Integer in = new Integer(i);
			
			if (hits.contains(in)) hl.attribs |= 1;
			if (misses.contains(in)) hl.attribs |= 2;
			
			v.add(hl);
			
			HighlightedLine lnum = new HighlightedLine();
			lnum.text = in + "  ";
			
			lnum.attribs = hl.attribs;
			
			lines.add(lnum);
		} // while
		
		r.close();
		
		lineData = new JList(v);
		lineNumbers = new JList(lines);
		
		lineData.setCellRenderer(highlightingRenderer);
		lineNumbers.setCellRenderer(highlightingRenderer);
		
		MouseListener ml = new MouseAdapter()
		{
			public void mouseClicked(MouseEvent e)
			{
				if (e.getClickCount() != 2 && !e.isPopupTrigger()) return;
				
				int index;
				if (e.getSource() == lineData) index = lineData.locationToIndex(e.getPoint());
				else index = lineNumbers.locationToIndex(e.getPoint());
				
				if (!linezeroshown) index++;
				
				Set les = (Set)entries.get(new Integer(index));
				if (les == null) return;
				
				StringBuffer report = new StringBuffer();
				
				Iterator it = les.iterator();
				while (it.hasNext())
				{
					TableEntry te = (TableEntry)it.next();
					report.append("Positions ");
					report.append(te.getStartPosition());
					report.append(" through ");
					report.append(te.getEndPosition());
					report.append(" in method ");
					report.append(te.getMethodName());
					report.append(", class ");
					String cname = te.getClassName();
					cname = cname.substring(cname.lastIndexOf(File.separatorChar) + 1);
					cname = cname.substring(0, cname.length() - 6);
					report.append(cname);
					report.append(", have");
					if (!te.wasExecuted()) report.append(" not");
					report.append(" been executed.\n");
				}
				
				JOptionPane.showMessageDialog(jScrollPane1, report);
			} // mouseClicked		
		}; // anonymous MouseListener
		lineData.addMouseListener(ml);
		lineNumbers.addMouseListener(ml);
		
		jPanel1.removeAll();
		
		jPanel1.add(lineData, BorderLayout.CENTER);
		jPanel1.add(lineNumbers, BorderLayout.WEST);
		
		jScrollPane1.setViewportView(jPanel1);
        summaryCheckBoxMenuItem.setSelected(false);
	} // fill
	
	// package visible for the Main gui only.
	boolean isStored()
	{
		return stored;
	} // isStored
	
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents()//GEN-BEGIN:initComponents
    {
        jScrollPane1 = new javax.swing.JScrollPane();
        jPanel1 = new javax.swing.JPanel();
        jMenuBar1 = new javax.swing.JMenuBar();
        fileMenu = new javax.swing.JMenu();
        closeItem = new javax.swing.JMenuItem();
        viewMenu = new javax.swing.JMenu();
        refreshItem = new javax.swing.JMenuItem();
        jSeparator1 = new javax.swing.JSeparator();
        summaryCheckBoxMenuItem = new javax.swing.JCheckBoxMenuItem();
        colorMenu = new javax.swing.JMenu();
        exItem = new javax.swing.JMenuItem();
        nexItem = new javax.swing.JMenuItem();
        mixItem = new javax.swing.JMenuItem();

        addWindowListener(new java.awt.event.WindowAdapter()
        {
            public void windowClosing(java.awt.event.WindowEvent evt)
            {
                exitForm(evt);
            }
        });

        jScrollPane1.setPreferredSize(new java.awt.Dimension(600, 440));
        jPanel1.setLayout(new java.awt.BorderLayout());

        jScrollPane1.setViewportView(jPanel1);

        getContentPane().add(jScrollPane1, java.awt.BorderLayout.CENTER);

        fileMenu.setText("File");
        closeItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_W, java.awt.event.InputEvent.CTRL_MASK));
        closeItem.setText("Close");
        closeItem.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                closeItemActionPerformed(evt);
            }
        });

        fileMenu.add(closeItem);
        jMenuBar1.add(fileMenu);
        viewMenu.setText("View");
        refreshItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_F5, 0));
        refreshItem.setText("Refresh View");
        refreshItem.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                refreshItemActionPerformed(evt);
            }
        });

        viewMenu.add(refreshItem);
        viewMenu.add(jSeparator1);
        summaryCheckBoxMenuItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_S, java.awt.event.InputEvent.CTRL_MASK));
        summaryCheckBoxMenuItem.setText("Summary View");
        summaryCheckBoxMenuItem.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                summaryCheckBoxMenuItemActionPerformed(evt);
            }
        });

        viewMenu.add(summaryCheckBoxMenuItem);
        jMenuBar1.add(viewMenu);
        colorMenu.setText("Colors");
        exItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_E, java.awt.event.InputEvent.CTRL_MASK));
        exItem.setBackground(exec);
        exItem.setText("Executed Color...");
        exItem.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                exItemActionPerformed(evt);
            }
        });

        colorMenu.add(exItem);
        nexItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_N, java.awt.event.InputEvent.CTRL_MASK));
        nexItem.setBackground(not);
        nexItem.setText("Not Executed Color...");
        nexItem.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                nexItemActionPerformed(evt);
            }
        });

        colorMenu.add(nexItem);
        mixItem.setAccelerator(javax.swing.KeyStroke.getKeyStroke(java.awt.event.KeyEvent.VK_M, java.awt.event.InputEvent.CTRL_MASK));
        mixItem.setBackground(mix);
        mixItem.setText("Mixed Color...");
        mixItem.addActionListener(new java.awt.event.ActionListener()
        {
            public void actionPerformed(java.awt.event.ActionEvent evt)
            {
                mixItemActionPerformed(evt);
            }
        });

        colorMenu.add(mixItem);
        jMenuBar1.add(colorMenu);
        setJMenuBar(jMenuBar1);

        pack();
    }//GEN-END:initComponents

    private void summaryCheckBoxMenuItemActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_summaryCheckBoxMenuItemActionPerformed
    {//GEN-HEADEREND:event_summaryCheckBoxMenuItemActionPerformed
        boolean selected = summaryCheckBoxMenuItem.isSelected();
        
        if (selected)
        {
            lineData.setCellRenderer(summaryRenderer);
            jScrollPane1.setViewportView(lineData);
            repaint();
        }
        else
        {
            lineData.setCellRenderer(highlightingRenderer);
            jPanel1.add(lineData, BorderLayout.CENTER);
            jScrollPane1.setViewportView(jPanel1);
            repaint();
        }
    }//GEN-LAST:event_summaryCheckBoxMenuItemActionPerformed

	private void mixItemActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_mixItemActionPerformed
	{//GEN-HEADEREND:event_mixItemActionPerformed
		Color c = JColorChooser.showDialog(this, "Select Color for Mixed Code", mix);
		if (c == null || mix.equals(c)) return;
		
		mix = c;
		
		mixItem.setBackground(mix);
		repaint();
		
		String propname = System.getProperty("user.home") + File.separator + ".gretelrc";
		
		Properties p = new Properties();
		try
		{
			InputStream is = new FileInputStream(propname);
			p.load(is);
			
			is.close();
		}
		catch (IOException e) {}
		
		p.setProperty(MIXED_COLOR_PROPERTY,
			"0x" + Integer.toHexString(mix.getRGB() & 0xffFFff));
		
		try
		{
			OutputStream os = new FileOutputStream(propname);
			p.store(os, "Gretel configuration file.");
			os.close();
		}
		catch (IOException e) {}
	}//GEN-LAST:event_mixItemActionPerformed

	private void nexItemActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_nexItemActionPerformed
	{//GEN-HEADEREND:event_nexItemActionPerformed
		Color c = JColorChooser.showDialog(this, "Select Color for Non-Executed Code", not);
		if (c == null || not.equals(c)) return;
		
		not = c;
		
		nexItem.setBackground(not);
		repaint();
		
		String propname = System.getProperty("user.home") + File.separator + ".gretelrc";
		
		Properties p = new Properties();
		try
		{
			InputStream is = new FileInputStream(propname);
			p.load(is);
			
			is.close();
		}
		catch (IOException e) {}
		
		p.setProperty(NOTEXECUTED_COLOR_PROPERTY,
			"0x" + Integer.toHexString(not.getRGB() & 0xffFFff));
		
		try
		{
			OutputStream os = new FileOutputStream(propname);
			p.store(os, "Gretel configuration file.");
			os.close();
		}
		catch (IOException e) {}
	}//GEN-LAST:event_nexItemActionPerformed

	private void refreshItemActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_refreshItemActionPerformed
	{//GEN-HEADEREND:event_refreshItemActionPerformed
		try
		{
			fill();
		}
		catch (IOException e)
		{
			JOptionPane.showMessageDialog(this, "IO Error attempting to refresh", "Error", JOptionPane.ERROR_MESSAGE);
		} 	
	}//GEN-LAST:event_refreshItemActionPerformed

	private void exItemActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_exItemActionPerformed
	{//GEN-HEADEREND:event_exItemActionPerformed
		Color c = JColorChooser.showDialog(this, "Select Color for Executed Code", exec);
		if (c == null || exec.equals(c)) return;
		
		exec = c;
		
		exItem.setBackground(exec);
		repaint();
		
		String propname = System.getProperty("user.home") + File.separator + ".gretelrc";
		
		Properties p = new Properties();
		try
		{
			InputStream is = new FileInputStream(propname);
			p.load(is);
			
			is.close();
		}
		catch (IOException e) {}
		
		p.setProperty(EXECUTED_COLOR_PROPERTY,
			"0x" + Integer.toHexString(exec.getRGB() & 0xffFFff));
		
		try
		{
			OutputStream os = new FileOutputStream(propname);
			p.store(os, "Gretel configuration file.");
			os.close();
		}
		catch (IOException e) {}
	}//GEN-LAST:event_exItemActionPerformed

	private void closeItemActionPerformed(java.awt.event.ActionEvent evt)//GEN-FIRST:event_closeItemActionPerformed
	{//GEN-HEADEREND:event_closeItemActionPerformed
		shutdown();
	}//GEN-LAST:event_closeItemActionPerformed

    private void exitForm(java.awt.event.WindowEvent evt) {//GEN-FIRST:event_exitForm
		shutdown();
    }//GEN-LAST:event_exitForm
	
	private void shutdown()
	{
		store();
		
		stored = true;
		
        if (quit)
		{
			System.exit(0);
		}
		else
		{
			dispose();
		}
	} // shutdown
	
	/**
	 * Left package visible to allow the Main gui to store an open window,
	 * thus causing the size to persist.
	 */
	void store()
	{
		String propname = System.getProperty("user.home") + File.separator + ".gretelrc";
		
		Properties p = new Properties();
		try
		{
			InputStream is = new FileInputStream(propname);
			p.load(is);
			
			is.close();
		}
		catch (IOException e) {}
		
		Rectangle r = getBounds();
		
		p.setProperty(TOP_BOUND_PROPERTY, String.valueOf(r.y));
		p.setProperty(LEFT_BOUND_PROPERTY, String.valueOf(r.x));
		p.setProperty(WIDTH_PROPERTY, String.valueOf(r.width));
		p.setProperty(HEIGHT_PROPERTY, String.valueOf(r.height));
		
		try
		{
			OutputStream os = new FileOutputStream(propname);
			p.store(os, "Gretel configuration file.");
			os.close();
		}
		catch (IOException e) {}
	} // store
	
	/**
	 * package visible method to correct bounds that are slightly off right
	 * after creation.
	 */
	void fixBounds()
	{
		setBounds(correctBounds);
	} // fixBounds
	
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JMenu fileMenu;
    private javax.swing.JPanel jPanel1;
    private javax.swing.JSeparator jSeparator1;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JMenuItem closeItem;
    private javax.swing.JMenuItem exItem;
    private javax.swing.JMenu viewMenu;
    private javax.swing.JMenu colorMenu;
    private javax.swing.JCheckBoxMenuItem summaryCheckBoxMenuItem;
    private javax.swing.JMenuItem nexItem;
    private javax.swing.JMenuItem mixItem;
    private javax.swing.JMenuItem refreshItem;
    private javax.swing.JMenuBar jMenuBar1;
    // End of variables declaration//GEN-END:variables

	/**
	 * A simple modification of the DefaultListCellRenderer to
	 * display the lines in tbe SourceViewer
	 *
	 * @author  chowells
	 * @version 
	 */
	private class HighlightingListCellRenderer extends DefaultListCellRenderer
	{
		public HighlightingListCellRenderer()
		{
			super();
			setBorder(noFocusBorder);
		} // HighlightingListCellRenderer
		
		public Component getListCellRendererComponent(
			JList list,
			Object value,
			int index,
			boolean isSelected,
			boolean cellHasFocus)
		{
			if (!(value instanceof HighlightedLine))
				throw new IllegalStateException("Internal error in SourceViewer");
			
			HighlightedLine hl = (HighlightedLine)value;
			
			setForeground(list.getForeground());
			
			switch (hl.attribs)
			{
				case 1:
					setBackground(exec);
					break;
				case 2:
					setBackground(not);
					break;
				case 3:
					setBackground(mix);
					break;
				default:
					setBackground(list.getBackground());
					break;
			} // switch
			
			setText(hl.text);
			setEnabled(list.isEnabled());
			setFont(list.getFont());
			
			return this;
		} // getListCellRendererComponent
		
	} // class HighlightingListCellRenderer
	
    private class SummaryListCellRenderer extends JComponent implements ListCellRenderer
    {
        private HighlightedLine hl;
        private JList drawingIn;
        
        public java.awt.Component getListCellRendererComponent(
            JList jList,
            Object value,
            int index,
            boolean isSelected,
            boolean cellHasFocus)
        {
			if (!(value instanceof HighlightedLine))
				throw new IllegalStateException("Internal error in SourceViewer");
			
			hl = (HighlightedLine)value;
            drawingIn = jList;
            return this;
        } // getListCellRenderer
        
        public void paint(java.awt.Graphics g)
        {
            // set the background to all white
            g.setColor(Color.white);
            g.fillRect(0, 0, getWidth(), getHeight());
            
            // create the row to draw in
            int [] row = new int[hl.text.length()];
            
            // get the color's int vals
            int background = Color.white.getRGB();
            int foreground;
            switch (hl.attribs)
            {
                case 1:
                    foreground = exec.getRGB();
                    break;
                case 2:
                    foreground = not.getRGB();
                    break;
                case 3:
                    foreground = mix.getRGB();
                    break;
                default:
                    foreground = drawingIn.getForeground().getRGB();
                    break;
            }
            
            // fill in row with the appropriate color:
            
            // non-whitespace found, for watching indentation
            boolean nwFound = false;
            for (int i = 0; i < hl.text.length(); i++)
            {
                if (!nwFound && Character.isWhitespace(hl.text.charAt(i)))
                {
                    row[i] = background;
                }
                else
                {
                    nwFound = true;
                    row[i] = foreground;
                }
            }
            
            MemoryImageSource mis = new MemoryImageSource(row.length, 1, row, 0, row.length);
            Image img = createImage(mis);
            
            g.drawImage(img, 0, 0, this);
            g.drawImage(img, 0, 1, this);
        } // paint();
        
        public Dimension getPreferredSize()
        {
            return new Dimension(hl.text.length(), 2);
        } // getPreferredSize
        
        // The following methods were all overridden to prevent them
        // from wasting time.  As they are called occasionally, but never
        // needed as part of a ListCellRenderer, they have been reimplemented
        // to do nothing.
        public void validate() {}
        public void revalidate() {}
        public void repaint(long tm, int x, int y, int width, int height) {}
        public void repaint(Rectangle r) {}
        public void firePropertyChange(String propertyName, Object oldValue, Object newValue) {}
        public void firePropertyChange(String propertyName, byte oldValue, byte newValue) {}
        public void firePropertyChange(String propertyName, char oldValue, char newValue) {}
        public void firePropertyChange(String propertyName, short oldValue, short newValue) {}
        public void firePropertyChange(String propertyName, int oldValue, int newValue) {}
        public void firePropertyChange(String propertyName, long oldValue, long newValue) {}
        public void firePropertyChange(String propertyName, float oldValue, float newValue) {}
        public void firePropertyChange(String propertyName, double oldValue, double newValue) {}
        public void firePropertyChange(String propertyName, boolean oldValue, boolean newValue) {}
    } // SummaryListCellRenderer
    
	private class HighlightedLine
	{
		public String text;
		public int attribs;
	} // HighlightedLine
	
} // class SourceViewer
